import time
from datetime import datetime
from threading import Thread

import pytest
from classic.scheduler import Scheduler


@pytest.fixture(scope='function')
def scheduler() -> Scheduler:
    scheduler = Scheduler()
    scheduler.start()

    yield scheduler

    scheduler.stop()


@pytest.fixture(scope='function')
def stopped_scheduler() -> Scheduler:
    return Scheduler()


def test__scheduler__with_delay(scheduler: Scheduler) -> None:
    result = None

    def task(a: int, b: int) -> None:
        nonlocal result
        result = a + b

    scheduler.with_delay(0.2, task, args=(1, ), kwargs=dict(b=2))

    time.sleep(0.18)
    assert result is None

    time.sleep(0.04)
    assert result == 3


def test__scheduler__cancel(scheduler: Scheduler) -> None:

    executed_tasks = []
    expected_tasks = []

    def task(task_name_: str) -> None:
        executed_tasks.append(task_name_)

    task_name = 'by_cron'
    scheduler.by_cron(
        '* * * * * *',
        task,
        args=(task_name, ),
        task_name=task_name,
    )
    scheduler.unshedule(task_name)
    time.sleep(0.045)
    scheduler.unshedule(task_name)

    task_name = 'with_delay'
    scheduler.with_delay(0.05, task, args=(task_name, ), task_name=task_name)
    time.sleep(0.045)
    scheduler.unshedule(task_name)

    task_name = 'by_period'
    scheduler.by_period(
        0.05,
        task,
        args=(task_name, ),
        task_name=task_name,
    )
    # Задача "by_period" выполняется сразу же
    expected_tasks.append(task_name)
    time.sleep(0.045)
    scheduler.unshedule(task_name)

    time.sleep(1.1)
    assert executed_tasks == expected_tasks


def test__scheduler__by_cron(scheduler: Scheduler) -> None:
    task_name = 'by_cron'
    all_results = [3, 5]
    results = []
    expected_results = []

    def task(a: int, b: int) -> None:
        a = results[-1] if results else a
        task_result = a + b
        results.append(task_result)

    scheduler.by_cron(
        '* * * * * *',
        task,
        args=(1, ),
        kwargs=dict(b=2),
        task_name=task_name,
    )
    for result in all_results:
        tick = 0.08
        now = datetime.utcnow()
        time_from = now.replace(microsecond=0)
        delta = 1 - (now - time_from).total_seconds()

        delay = max(delta - tick, 0)
        time.sleep(delay)
        assert results == expected_results

        time.sleep(delta + tick)
        expected_results.append(result)
        assert results == expected_results

    scheduler.unshedule(task_name)


def test__scheduler__by_period(scheduler: Scheduler) -> None:
    task_name = 'by_period'
    all_results = [3, 5, 7]
    results = []
    expected_results = []

    def task(a: int, b: int) -> None:
        a = results[-1] if results else a
        task_result = a + b
        results.append(task_result)

    scheduler.by_period(
        0.2,
        task,
        args=(1, ),
        kwargs=dict(b=2),
        task_name=task_name,
    )

    time.sleep(0.02)
    for result in all_results:
        expected_results.append(result)
        assert results == expected_results

        time.sleep(0.17)
        assert results == expected_results
        time.sleep(0.03)

    scheduler.unshedule(task_name)


def test__scheduler__run(stopped_scheduler: Scheduler) -> None:
    result = False

    def run() -> None:
        stopped_scheduler.with_delay(0.05, task)
        time.sleep(0.03)
        stopped_scheduler.stop()

    def task() -> None:
        nonlocal result
        result = True

    thread = Thread(target=run)
    thread.start()

    stopped_scheduler.loop()
    time.sleep(0.05)
    assert not thread.is_alive()
    assert not result



import logging
from abc import ABC, abstractmethod
from datetime import datetime, timedelta, timezone
from typing import Any, Callable, Optional

from croniter import croniter


class Task(ABC):
    """
    Задача, выполняемая в потоке планировщика.
    """

    def __init__(
        self,
        job: Callable[[...], None],
        args: tuple = None,
        kwargs: dict = None,
        name: Optional[str] = None,
    ) -> None:
        self._job = job
        self._args = args or ()
        self._kwargs = kwargs or {}
        self._name = name
        self._next_run_time = None  # время следующего запуска задачи

    @property
    def name(self) -> str:
        return self._name

    @property
    def next_run_time(self) -> datetime:
        """
        Свойство возвращающее время следующего запуска.

        Returns:
            datetime: Время следующего запуска.
        """
        return self._next_run_time

    @abstractmethod
    def set_next_run_time(self) -> None:
        """
        Метод рассчитывающий у устанавливающий время следующего запуска.
        Переопределяется в дочерних классах.
        """

    def run_job(self) -> None:
        """
        Вызывает переданный Callable объект со всем аргументами
        в потоке планировщика.
        """
        logging.info(
            'Task [%s] started with schedule [%s]',
            self._name,
            self._next_run_time,
        )

        try:
            self._job(*self._args, **self._kwargs)
        except Exception as ex:
            logging.exception(
                'Unexpected error occurred in task [%s]: "%s".',
                self._name,
                ex,
            )
        else:
            logging.info('Task completed [%s]', self._name)

    def __lt__(self, other: 'Task') -> bool:
        return self._next_run_time < other._next_run_time

    def __gt__(self, other: 'Task') -> bool:
        return self._next_run_time > other._next_run_time


class OneTimeTask(Task):
    """
    Одноразовая задача, которая будет выполнена ровно один раз.
    """

    def __init__(self, date: datetime, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._next_run_time = date

    def set_next_run_time(self) -> None:
        self._next_run_time = None


class CronTask(Task):
    """
    Задача, которая выполняется по CRON расписанию.
    """

    def __init__(self, schedule: str, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._schedule = croniter(schedule, datetime.now(timezone.utc))
        self._next_run_time = self._schedule.get_next(datetime)

    def set_next_run_time(self) -> None:
        self._next_run_time = self._schedule.get_next(datetime)


class PeriodicTask(Task):
    """
    Задача, которая будет выполняться периодически.
    """

    def __init__(self, period: timedelta, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._period = period
        self._next_run_time = datetime.now(timezone.utc)

    def set_next_run_time(self) -> None:
        self._next_run_time = datetime.now(timezone.utc) + self._period



import heapq
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from queue import Empty, Queue
from threading import Thread
from typing import Any, Callable, Union, Optional

from .task import CronTask, OneTimeTask, PeriodicTask, Task


Stop = object()


@dataclass
class CancelTask:
    """
    Команда отмены запланированной задачи.
    """
    task_name: str


class Scheduler:
    """
    Планировщик задач. Вызывает переданные ему Callable объекты в соответствии
    с расписанием.
    """
    _stopped: bool
    _thread: Optional[Thread]
    _thread_pool: Optional[ThreadPoolExecutor]

    def __init__(self, workers_num: int = 1) -> None:
        super().__init__()
        self._thread = None
        self._tasks = []
        self._inbox = Queue()
        if workers_num:
            self._thread_pool = ThreadPoolExecutor(max_workers=workers_num)
        else:
            self._thread_pool = None

    def _calc_next_task_execute_time(self) -> float:
        """
        Возвращает время ожидания выполнения следующей задачи.

        Returns:
            float: тайм-аут ожидания задачи.
        """
        if len(self._tasks) < 1:
            return 60

        dt_next_task = self._tasks[0].next_run_time
        dt_current = datetime.now(timezone.utc)
        if dt_next_task < dt_current:
            return 0
        else:
            return (dt_next_task - dt_current).total_seconds()

    def _schedule(self, task: Task) -> None:
        heapq.heappush(self._tasks, task)

    def _cancel(self, task_name: str) -> None:
        """
        Отменяет выполнение запланированной задачи.

        Args:
            task_name (str): Имя задачи для отмены.
        """
        tasks = [t for t in self._tasks if t.name != task_name]
        if len(self._tasks) != len(tasks):
            heapq.heapify(tasks)
            self._tasks = tasks

    def _execute_task(self, task: Task) -> None:
        """
        Выполняет текущую задачу.

        Args:
            task (Task): Задача, которую необходимо выполнить.
        """
        if self._thread_pool:
            self._thread_pool.submit(task.run_job)
        else:
            task.run_job()

    def _stop(self) -> None:
        self._stopped = True

    def run(self) -> None:
        """
        Главный цикл выполнения планировщика.
        """
        self._stopped = False
        while not self._stopped:
            try:
                timeout = self._calc_next_task_execute_time()
                if timeout > 0:
                    try:
                        command = self._inbox.get(block=True, timeout=timeout)
                    except Empty:
                        continue

                    if isinstance(command, Task):
                        self._schedule(command)
                    elif isinstance(command, CancelTask):
                        self._cancel(command.task_name)
                    elif command is Stop:
                        self._stop()
                else:
                    task: Task = heapq.heappop(self._tasks)
                    task.set_next_run_time()
                    if task.next_run_time:
                        self._schedule(task)
                    self._execute_task(task)
            except KeyboardInterrupt:
                self._stop()

        if self._thread_pool:
            self._thread_pool.shutdown(wait=True, cancel_futures=True)

    def with_delay(
        self,
        delay: Union[timedelta, float],
        job: Callable[[], None],
        args: tuple = None,
        kwargs: dict = None,
        task_name: str = None,
    ) -> None:
        """
        Добавляет отложенную задачу в планировщик.
        Задача будет выполнена ровно один раз.

        Args:
            delay (Union[timedelta, float]): Отсрочка выполнения задачи.
            job (Callable[[], Any]): Вызываемая в задаче функция.
            args (tuple, optional): Аргументы. По умолчанию None.
            kwargs (dict, optional): Аргументы. По умолчанию None.
            task_name (str, optional): Имя таски. По умолчанию None.
        """
        if not isinstance(delay, timedelta):
            delay = timedelta(seconds=delay)

        self._inbox.put(OneTimeTask(
            date=datetime.now(timezone.utc) + delay,
            job=job,
            args=args,
            kwargs=kwargs,
            name=task_name,
        ))

    def by_cron(
        self,
        schedule: str,
        job: Callable[[], Any],
        args: tuple = None,
        kwargs: dict = None,
        task_name: str = None,
    ) -> None:
        """
        Добавляет задачу которая будет выполнятся по CRON расписанию.

        Args:
            schedule (str): Расписание задачи в CRON формате.
            job (Callable[[], Any]): Вызываемая в задаче функция.
            args (tuple, optional): Аргументы. По умолчанию None.
            kwargs (dict, optional): Аргументы. По умолчанию None.
            task_name (str, optional): Имя таски. По умолчанию None.
        """
        self._inbox.put(CronTask(
            schedule=schedule,
            job=job,
            args=args,
            kwargs=kwargs,
            name=task_name,
        ))

    def by_period(
        self,
        period: Union[timedelta, float],
        job: Callable[[], Any],
        args: tuple = None,
        kwargs: dict = None,
        task_name: str = None,
    ) -> None:
        """
        Добавляет задачу которая будет выполняется периодически через
        указанный интервал времени.

        Args:
            period (Union[timedelta, float]): Период выполнения задачи.
            job (Callable[[], Any]): Вызываемая в задаче функция.
            args (tuple, optional): Аргументы. По умолчанию None.
            kwargs (dict, optional): Аргументы. По умолчанию None.
            task_name (_type_, optional): Имя таски. По умолчанию None.
        """
        if not isinstance(period, timedelta):
            period = timedelta(seconds=period)

        self._inbox.put(PeriodicTask(
            period=period,
            job=job,
            args=args,
            kwargs=kwargs,
            name=task_name,
        ))

    def cancel(self, task_name: str) -> None:
        """
        Отменяет выполнение запланированной задачи.

        Args:
            task_name (str): Имя задачи для отмены.
        """
        cancellation = CancelTask(task_name=task_name)
        self._inbox.put(cancellation)

    def start(self) -> None:
        """
        Запускает выполнение планировщика.
        """
        if not self._thread or not self._thread.is_alive():
            # TODO: Без deamon=True потоки не останавливались до завершения
            # выполнения задачи. Нужно найти аналогичное решение для
            # корректной остановки потока.
            self._thread = Thread(target=self.run, daemon=True)
            self._thread.start()

    def stop(self, block: bool = True, timeout: float = None) -> None:
        """
        Завершает выполнение планировщика.
        """
        if self._thread and self._thread.is_alive():
            self._inbox.put(Stop)
            if block:
                self._thread.join(timeout)

    def join(self, timeout: float = None):
        if self._thread and self._thread.is_alive():
            self._thread.join(timeout)
